#!/usr/bin/env python3
"""Un-categorized feature of `pymytools`."""

from pathlib import Path
import simpleaudio as sa
from pymytools.logger import markup, console, SUPPORTING_COLORS
import time


class Me:
    """Memes about me. :D"""

    _path_to_life: Path = Path(
        Path(__file__).parent.parent, "assets/sounds/life_is.wav"
    )

    @property
    def how_is_your_life(self) -> None:

        wave_obj = sa.WaveObject.from_wave_file(self._path_to_life.as_posix())
        play_obj = wave_obj.play()
        play_obj.wait_done()
        
    @property
    def python_is(self) -> None:
        
        for c in SUPPORTING_COLORS:
            console.print("\t🔥 " + markup("STATICALLY TYPED", c, "bold")+" 🔥", end="\r")
            time.sleep(0.5)
        print("")


Sun = Me()



#!/usr/bin/env python3
"""Returning list of indices for a various purposes."""
from itertools import combinations_with_replacement


def tensor_idx(dim: int) -> list[tuple[int, int]]:
    """Return the list of indices for the Cartesian tensor.

    Example:

        >>> c_tensor(3)
        [(0, 0), (0, 1), (0, 2), (1, 1), (1, 2), (2, 2)]
        >>> c_tensor(2)
        [(0, 0), (0, 1), (1, 1)]
        >>> c_tensor(1)
        [(0, 0)]


    Args:
        dim (int): Dimension of the tensor.

    Returns:
        list[tuple[int, int]]: List of indices.
    """

    assert dim in [1, 2, 3], "Only 1, 2, and 3 dimensions are supported."

    return list(combinations_with_replacement(range(dim), 2))



#!/usr/bin/env python3
"""Logging tools"""
import logging

from rich.console import Console
from rich.console import JustifyMethod
from rich.console import RenderableType
from rich.logging import RichHandler
from rich.style import StyleType
from rich.table import Table

console = Console()

from dataclasses import dataclass, field
import time

logging.basicConfig(
    level="NOTSET",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(markup=True)],
)

log = logging.getLogger("rich")

SUPPORTING_COLORS: list[str] = [
    "black",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "bright_black",
    "bright_red",
    "bright_green",
    "bright_yellow",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_white",
]


@dataclass
class Report:
    """Report the dictionary type data using the rich table."""

    title: str
    content: dict[str, list[RenderableType]]
    justify: JustifyMethod | list[JustifyMethod] | None = "center"
    style: StyleType | list[StyleType] | None = "cyan"

    def __post_init__(self):
        self.table = Table(title=self.title)

    def display(self) -> None:
        """Print out the rich table."""

        # Construct header
        for idx, (k, _) in enumerate(self.content.items()):
            if isinstance(self.justify, list):
                justify = self.justify[idx]
            elif self.justify is None:
                justify = "center"
            else:
                justify = self.justify

            if isinstance(self.style, list):
                style = self.style[idx]
            elif self.justify is None:
                style = "cyan"
            else:
                style = self.style

            self.table.add_column(k, justify=justify, style=style)

        render: list[list[str]] = []

        for _, values in self.content.items():
            # Rich only accept string as a renderable object
            render.append([str(v) for v in values])

        renderables = [list(i) for i in zip(*render)]
        for r in renderables:
            self.table.add_row(*r)

        console.print(self.table)


@dataclass
class Timer:
    """Pystops timer class. You can import this module with name `Timer`.
    This class is intended to compute elapsed time of computation with name flag.

    >>> Timer
    >>> Timer.start("sim")
    >>> time.sleep(1) # Any computation
    >>> Timer.end("sim")
    >>> Timer.elapsed("sim)
    ~ 1 # in second

    """

    clock: dict[str, dict[str, float]] = field(default_factory=dict)

    def start(self, name: str) -> None:
        """Start timer with name"""

        now = time.perf_counter()

        if name in self.clock:
            self.clock[name]["start"] = now
        else:
            self.clock[name] = {"start": now, "elapsed": 0.0}
            self.clock[name]["n_calls"] = 0

    def end(self, name: str) -> None:
        """End timer with name"""

        try:
            now = time.perf_counter()
            elapsed = now - self.clock[name]["start"]
            self.clock[name]["elapsed"] += elapsed
            self.clock[name]["n_calls"] += 1
        except KeyError:
            raise RuntimeError("Timer: timer is not started!")

    def elapsed(self, name: str) -> float:
        """Return elapsed time with name"""

        return self.clock[name]["elapsed"]

    def display(self) -> None:
        table = Table(title="Timer Report")
        table.add_column("Name", justify="center", style="cyan")
        table.add_column(r"Elapsed time \[s]", justify="center", style="green")
        table.add_column("#Calls", justify="center", style="magenta")

        for k, v in self.clock.items():
            table.add_row(k, f"{v['elapsed']:.5f}", f"{v['n_calls']}")

        console.print(table)


timer = Timer()


def markup(msg: str, color: str | None = None, style: str | None = None) -> str:
    """Generate rich markup string.

    Args:
        color (str): Color of the text. Color should be one of basic terminal 16 colors, ["black", "red", "green", "yellow", "blue", "magenta", "cyan", "white", "bright_black", "bright_red", "bright_green", "bright_yellow", "bright_blue", "bright_magenta", "bright_cyan", "bright_white"]
        msg (str): Message to be printed.
        style (str): Style of the text. See rich style list. One of "bold", "italic", "underline", "strike".
    """

    tag = "["

    if style is None:
        tag += ""
    else:
        assert style in ["bold", "italic", "underline", "strike", "blink"] or style in [
            "b",
            "i",
            "u",
            "s",
            "bk",
        ]
        tag += style

    if color is None:
        tag += ""
    else:
        assert color in SUPPORTING_COLORS
        if len(tag) > 1:
            tag += " " + color
        else:
            tag += color

    tag += "]"

    return f"{tag}{msg}[/{tag[1:]}"


def rich_report(num_calls: int, elapsed_time: float, name: str) -> Table:
    """Report timer using rich table."""

    table = Table(title="Timer Report")
    table.add_column("Name", justify="center", style="cyan")
    table.add_column("Number of calls [-]", justify="center", style="magenta")
    table.add_column(r"Elapsed time \[s]", justify="center", style="green")

    table.add_row(name, f"{num_calls}", f"{elapsed_time:.5f}")

    return table


def draw_rule(width: int = 80, character: str = "=", color: str = "green") -> str:
    """Draw a rule for a given width and color."""

    if character == "-":
        character = chr(9472)
    elif character == "=":
        character = chr(9552)
    else:
        msg = f"Invalid line style ({character})! Should be one of (-) or (=)."
        raise ValueError(msg)

    rule = ""

    for _ in range(width):
        rule += character

    console.print(markup(rule, color=color, style="bold"))

    return rule



#!/usr/bin/env python
"""Pretty progress bar using rich and tqdm."""
from time import sleep

from rich.progress import BarColumn
from rich.progress import filesize
from rich.progress import Progress as RichProgress
from rich.progress import ProgressColumn
from rich.progress import TimeElapsedColumn
from rich.progress import TimeRemainingColumn
from rich.text import Text
from tqdm.std import tqdm


class FractionColumn(ProgressColumn):
    """Renders completed/total, e.g. '0.5/2.3 G'."""

    def __init__(self, unit_divisor=1000):
        self.unit_divisor = unit_divisor
        super().__init__()

    def render(self, task):
        """Calculate common unit for completed and total."""
        completed = int(task.completed)
        total = int(task.total)  # type: ignore
        unit, suffix = filesize.pick_unit_and_suffix(total, [""], 1)

        precision = 0 if unit == 1 else 1
        return Text(
            f"{completed/unit:,.{precision}f}/{total/unit:,.{precision}f} {suffix}",
            style="progress.download",
        )


class RateColumn(ProgressColumn):
    """Renders human readable transfer speed."""

    def __init__(self, unit="", unit_divisor=1000):
        self.unit = unit
        self.unit_divisor = unit_divisor
        super().__init__()

    def render(self, task):
        """Show data transfer speed."""
        speed = task.speed

        if speed is None:
            return Text(f"? {self.unit}/s", style="progress.data.speed")
        else:
            unit, suffix = filesize.pick_unit_and_suffix(speed, [""], 1)  # type: ignore
        precision = 0 if unit == 1 else 1

        return Text(
            f"{speed/unit:,.{precision}f} {suffix}{self.unit}/s",
            style="progress.data.speed",
        )


rich_progress_config = (
    "[bold yellow][progress.description]{task.description}[/bold yellow]"
    "[progress.percentage]{task.percentage:>4.0f}%",
    BarColumn(bar_width=30),
    FractionColumn(unit_divisor=1000),
    "[",
    TimeElapsedColumn(),
    "<",
    TimeRemainingColumn(),
    ",",
    RateColumn(unit="it", unit_divisor=1000),
    "]",
)


class ProgressBar(tqdm):
    """Custom progress bar inspired by `tqdm.rich`.

    Example:
        >>> tasks = range(20)
        >>> pbar = ProgressBar(tasks)
        >>> for idx in pbar:
                sleep(0.1)
                pbar.set_description(f"Processing")
        Processing 100% |████████████████████████████████� 20/20 [00:02<00:00, 10 it/s]
        # Or you can try this example via `python -m pystops_tool.progress`

    """

    def __init__(self, *args, **kwargs):
        """
        This class accepts the following parameters *in addition* to
        the parameters accepted by `tqdm`.

        Args:
            progress (tuple, optional): arguments for `rich.progress.Progress()`.
        """
        kwargs = kwargs.copy()
        kwargs["gui"] = True
        kwargs["disable"] = False
        progress = kwargs.pop("progress", None)

        super().__init__(*args, **kwargs)

        d = self.format_dict

        # Allows custom config
        if progress is None:
            progress = rich_progress_config

        self._prog = RichProgress(*progress, transient=not self.leave)
        self._prog.__enter__()
        self._task_id = self._prog.add_task(self.desc or "", **d)

    def close(self):
        super().close()
        self._prog.__exit__(None, None, None)

    def clear(self, *_, **__):
        pass

    def display(self, *_, **__):
        self._prog.update(self._task_id, completed=self.n, description=self.desc)

    def reset(self, total: int | float | None = None):
        """
        Resets to 0 iterations for repeated use.

        Args:
            total (int or float, optional): Total to use for the new bar.
        """
        self._prog.reset(self._task_id, total=total)
        super().reset(total=total)


if __name__ == "__main__":
    """Display some examples."""

    tasks = range(20)
    pbar = ProgressBar(tasks)

    for idx in pbar:
        sleep(0.1)
        pbar.set_description(f"Processing")



#!/usr/bin/env python3
"""Torch wrapper for the scipy.special functions.

* This functions are much slower than scipy.special functions. However, we need to use it for `torch.vmap` utilization.

Performance:
    - n=100000
        ┏━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┓
        ┃     Name     ┃   Elapsed time (s)   ┃
        ┡━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━┩
        │ scipy_ellipe │ 0.001932209124788642 │
        │ scipy_ellipk │ 0.025676249992102385 │
        │ vamp_ellipe  │ 0.05203570798039436  │
        │ vmap_ellipk  │ 0.08591929194517434  │
        └──────────────┴──────────────────────┘

Note on `t_ellipk` and `t_ellipe` (`torch.vmap` is not working for this function.):

- Since torch doesn't provide `ellipk` and `ellipe` functions, we need to wrap around `scipy.special` functions. This is necessary to utilize GPU backend.
- Here, we convert input tensor to cpu and return as `torch.device` Tensor.
- Performance is just comparable to direct `scipy` function call with `cpu`. But using `gpu` backend is slightly worse.
    - I guess this is due to the copy of memory while it converts `gpu` tensor to `cpu` tensor.
    - Also, the actual computation is in `cpu` so it is expected.
    - Elliptic integral Performance, n=1000000
        ┏━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━┓
        ┃     Name     ┃ Elapsed time (s) ┃
        ┡━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━┩
        │ scipy_ellipe │     0.00806      │
        │ scipy_ellipk │     0.00845      │
        │ torch_ellipe │     0.00824      │
        │ torch_ellipk │     0.00841      │
        │ torch (gpu)  │     0.01007      │
        └──────────────┴──────────────────┘
"""
from pymytools.special._ellipe import ellipe
from pymytools.special._ellipe import t_ellipe
from pymytools.special._ellipk import ellipk
from pymytools.special._ellipk import t_ellipk


__all__ = ["ellipk", "ellipe", "t_ellipk", "t_ellipe"]



#!/usr/bin/env python3
"""
Store all relevant constants. It contains

* Electron mass
* Proton mass
* Neutron mass
* Atomic mass
* Speed of light
* Magnetic constant
* Vacuum permittivity
* Boltzmann constant
* Converting factor from eV to K
* Elementary charge
* Converting factor from K to eV
* Pi

"""
import numpy as _np
from scipy import constants as _constants

ME: float = _constants.m_e
"""Electron mass [kg]"""
MP: float = _constants.m_p
"""Proton mass [kg]"""
MN: float = _constants.m_n
"""Neutron mass [kg]"""
AMU: float = _constants.m_u
"""Atomic mass [kg]"""
C: float = _constants.c
"""Speed of light in vacuum [m/s]"""
MU0: float = _constants.mu_0
"""Magnetic constant [H/m]"""
EPS0: float = _constants.epsilon_0
"""Vacuum permittivity [F/m]"""
KB: float = _constants.k
"""Boltzmann constant [J/K]"""
eVtoK: float = _constants.value("electron volt-kelvin relationship")
"""Converting factor from eV to K"""
QE: float = _constants.value("elementary charge")
"""Elementary charge 1.60217662 × 10-19 [C]"""
KtoeV: float = 1.0 / eVtoK
"""Converting factor from K to eV"""
PI: float = _np.pi
"""PI"""



#!/usr/bin/env python3
"""Diagnotstics for pystops project.

It contains
    - Save gridded data
    - Save discrete (particle) data
    - Tensorboard

Note that we assume all data are in the form of `torch.Tensor`.
"""
import glob
import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TypedDict

import h5py
import numpy as np
import pandas as pd
import pyevtk
import torch
from pymyplot import Figure
from torch import Tensor
from torch.utils.tensorboard.writer import SummaryWriter
from vtk import vtkXMLUnstructuredGridReader
from vtk.util.numpy_support import vtk_to_numpy  # type: ignore
from vtkmodules.vtkIOXML import vtkXMLImageDataReader

from pymytools.logger import logging
from pymytools.logger import markup

PATHLIKE = str | Path | os.PathLike
"""Path types. Either `str`, `pathlib.Path` or `os.PathLike`"""

DTYPE_T2N = {
    torch.float32: np.float32,
    torch.float64: np.float64,
    torch.int32: np.int32,
    torch.int64: np.int64,
}
"""Datatype converter from torch.dtype to numpy.dtype"""


class VTILoaded(TypedDict):
    """Loaded data from VTK file."""

    grid: tuple[Tensor, ...]
    """Grid information"""
    data: dict[str, Tensor]
    """Retrived data"""


class VTULoaded(TypedDict):
    """Loaded data from VTK file."""

    pos: Tensor
    "Particle position"
    vel: Tensor
    "particle velocity"


@dataclass
class DataLoader:
    """Load data from path.

    It supports following file types:
        - VTK (.vti, .vtu)
        - CSV (.csv)
        - HDF5 (.h5)

    Note:
        - Reading `vtk` format is not straightforward. Also, `pyevtk` does not support. Therefore, use `vtk` module directly.

    Args:
        dtype (torch.dtype): Data type of loaded data. Default is `torch.float64`.
        device (torch.device): Device of loaded data. Default is `torch.device("cpu")`.
    """

    dtype: torch.dtype = torch.float64
    device: torch.device = torch.device("cpu")

    def read_state_dict(
        self, architecture: torch.nn.Module, model_path: PATHLIKE
    ) -> torch.nn.Module:
        """Load state dict of the nn architecture."""

        state_dict_loc = Path(model_path).as_posix()

        architecture.load_state_dict(
            torch.load(state_dict_loc, map_location=self.device)
        )

        return architecture

    def read_vti(self, file_path: str, key: str | list[str] | None = None) -> VTILoaded:
        """Read VTK image data.

        Example:
            >>> loader = DataLoader()
            >>> res = loader.read_vtk("test.vti", key="test")
            # If a list of keys is given, it returns all data corresponding to the keys.
            # If None is given as the key, it returns all data.
            >>> res["grid"]
            tuple(Tensor, ...) # len(data["grid"]) is based on mesh dimension
            >>> res["data"]
            {key: Tensor}
            # res["grid"][0].shape == res["data"][key].shape

        Args:
            file_path (str): Path to load data. Note that `vtk` only accept string.
            key (str | list[str] | None): Key to retrive data. If `None` is given, it returns all data. Default is `None`.

        Returns:
            VTILoaded: loaded data, which is a dictionary with two keys: `grid` and `data`.
        """

        path = Path(file_path)
        if is_file(path) is False:
            raise FileNotFoundError(f"File {path} does not exist.")

        reader = vtkXMLImageDataReader()
        reader.SetFileName(path.as_posix())
        reader.Update()
        cell = reader.GetOutput()
        n_components = cell.GetCellData().GetNumberOfComponents()

        if isinstance(key, str):
            key = [key]
        elif key is None:
            key = [cell.GetCellData().GetArrayName(i) for i in range(n_components)]
        elif isinstance(key, list):
            assert all(
                isinstance(i, str) for i in key
            ), "Dataloader: some or all of keys are not str."
        else:
            raise TypeError("Dataloader: wrong type for key.")

        dx = cell.GetSpacing()
        origin = cell.GetOrigin()
        extent = cell.GetExtent()
        nx = [extent[2 * i] + extent[2 * i + 1] for i in range(3)]

        # Determine the dimension of the data
        # We put 1 grid point with the size of dx for dummy dimension
        data_dim = 3
        dim_check = 0
        for n in nx:
            if n == 1:
                dim_check += 1
                break
        data_dim -= dim_check

        cell_data: dict[str, Tensor] = {}

        # Retrive VTK cell data
        for k in key:
            # Convert VTK data to numpy array using key
            cell_val = vtk_to_numpy(cell.GetCellData().GetScalars(k))
            cell_val = cell_val.reshape(*nx[:data_dim]).T
            cell_dtype = cell_val.dtype

            # Convert numpy array to torch.Tensor
            cell_val = torch.tensor(cell_val, dtype=self.dtype, device=self.device)
            if cell_dtype == np.int32:
                cell_val = cell_val.gt(0.0)

            cell_data[k] = cell_val

        grid = [
            torch.linspace(
                origin[i] + dx[i] / 2,
                origin[i] + dx[i] / 2 + dx[i] * (nx[i] - 1),
                nx[i],
                dtype=self.dtype,
                device=self.device,
            )
            for i in range(data_dim)
        ]
        mesh = torch.meshgrid(grid, indexing="ij")

        return {"grid": mesh, "data": cell_data}

    def read_vtu(self, file_path: str) -> VTULoaded:
        """Read VTK point data (unstructured grid).
        In my use case, `.vtu` file contains particle data usually either particle position or velocity. Especially the velocity information follows a convention of its identifier `U`, `V`, `W` for each dimension, therefore, we will return concatenated `Tensor` which is the data corresponding to `UVW` convention (lower case accepted) and has a dimension of `num_particle x 3`. Other than this keyword will be seprately loaded as single column (vector) data.

        Note:
            - Since this method only works for particle position and velocity, no key is required.

        Example:
            >>> dl = DataLoader()
            >>> data = dl.read_vtu("test.vtu")
            >>> data["pos"]
            Tensor # shape: num_particle x 3
            >>> data["vel"]
            Tensor # shape: num_particle x 3

        Args:
            file_path (str): Path to load data. Note that `vtk` only accept string.

        Returns:
            VTULoaded: Loaded data contains, which is a dictionary with two keys: `pos` and `vel`.
        """
        path = Path(file_path)

        if is_file(path) is False:
            raise FileNotFoundError(f"File {path} does not exist.")

        reader = vtkXMLUnstructuredGridReader()
        reader.SetFileName(path.as_posix())
        reader.Update()
        point = reader.GetOutput()
        n_components = point.GetPointData().GetNumberOfComponents()

        # Key for particle data
        key = [point.GetPointData().GetArrayName(i) for i in range(n_components)]

        concat: list[Tensor] = []

        # Retrive VTK point data
        for k in key:
            # Convert VTK data to numpy array using key
            point_val = vtk_to_numpy(point.GetPointData().GetScalars(k))

            # Convert numpy array to torch.Tensor
            concat.append(torch.tensor(point_val, dtype=self.dtype, device=self.device))

        # Save VTK velocity data
        data_vel = torch.vstack(concat).T

        # Retrive VTK position data
        pos_val = vtk_to_numpy(point.GetPoints().GetData())
        data_pos = torch.tensor(pos_val, dtype=self.dtype, device=self.device)

        return {"pos": data_pos, "vel": data_vel}

    def read_csv(
        self, path: PATHLIKE, key: str | list[str] | None = None
    ) -> dict[str, Tensor]:
        """Read csv data using `pandas.read_csv`.

        Example:
            >>> ds = DataLoader()
            >>> res = ds.read_csv("test.csv", "x")
            >>> res
            {"x": Tensor}
            >>> res = ds.read_csv("test.csv", ["x", "y"])
            >>> res
            {"x": Tensor, "y": Tensor}

        Args:
            path (PATHLIKE): Path to load data.
            key (str or list[str], optional): Key for the loaded data. If None, all data will be loaded. Defaults to None

        Returns:
            dict[str, Tensor]: Loaded data.
        """

        if is_file(path) is False:
            raise FileNotFoundError(f"File {path} does not exist.")

        df = pd.read_csv(path)
        if isinstance(key, str):
            key = [key]
        elif key is None:
            key = list(df.columns)
        elif isinstance(key, list):
            assert all(
                [isinstance(k, str) for k in key]
            ), "key must be str or list[str]."
        else:
            raise TypeError("Dataloader: wrong type for key.")

        data_csv: dict[str, Tensor] = {}

        for k in key:
            data_csv[k] = torch.tensor(
                df[k].values, dtype=self.dtype, device=self.device
            )

        return data_csv

    def read_hdf5(
        self, path: PATHLIKE, key: str | list[str] | None = None
    ) -> dict[str, Tensor]:
        """Read hdf5 file. Loaded data will be

        Args:
            path (PATHLIKE): Path to load data.
            addr (str or list[str], optional): address for the loaded data (group name).

        Returns:
            Tensor: Loaded data.
        """

        if is_file(path) is False:
            raise FileNotFoundError(f"File {path} does not exist.")

        with h5py.File(path, "r") as f_h5:
            # If there is no subgroup, the keys are the name in datasets
            # If there are subgroups, the keys are the name of the subgroups
            all_keys = get_hdf5_dataset_keys(f_h5)

            if isinstance(key, str):
                key = [key]
            elif isinstance(key, list):
                assert all(
                    isinstance(k, str) for k in key
                ), "Dataloader: all entry of the given list of key must be str."
            elif key is None:
                key = [ak.rsplit("/", 1)[-1] for ak in all_keys]

            # Key to group + key
            key_to_get: list[str] = []
            for k in key:
                for ak in all_keys:
                    if k == ak.rsplit("/", 1)[-1]:
                        key_to_get.append(ak)

            assert len(key_to_get) == len(
                key
            ), f"Dataloader: some of given keys don't exist in hdf data set. (given: {key}, exist: {key_to_get})"

            data_hdf5: dict[str, Tensor] = {}

            for k in key_to_get:
                data = f_h5[k]

                assert isinstance(data, h5py.Dataset)

                data_hdf5[k.rsplit("/", 1)[-1]] = torch.tensor(
                    data[:], dtype=self.dtype, device=self.device
                )

            return data_hdf5


def get_hdf5_dataset_keys(f: h5py.File) -> list[str]:
    keys = []
    f.visit(lambda key: keys.append(key) if isinstance(f[key], h5py.Dataset) else None)
    if None in keys:
        raise KeyError("Dataloader: no keys exists in the given hdf5 dataset.")
    return keys


@dataclass
class DataSaver:
    """DataSaver class

    Args:
        save_dir (str): Directory to save data.
    """

    save_dir: str
    """Save directory"""

    def __post_init__(self) -> None:
        # Validate save directory. If not exist, create it.
        is_dir(self.save_dir, create=True)

    def save_model(
        self, model: torch.nn.Module, model_name: str | None = None, verbose: int = 0
    ):
        """Save the trained model state dict.

        Note:
            - The model is always saved as `model_name.pth` in the save directory.

        Args:
            model (torch.nn.Module): trained model
            model_name (str, optional): Model name. Defaults to "".
            verbose (int, optional): Verbosity level. Defaults to 0.
        """

        if model_name is None:
            model_name = "model.pth"
        else:
            model_name += ".pth"

        if verbose > 0:
            logging.info(markup(f"Saving model to {self.save_dir}", "blue", "bold"))

        torch.save(model.state_dict(), Path(self.save_dir, model_name))

    def save_hdf5(
        self, data: dict[str, Tensor] | dict[str, dict[str, Tensor]], file_name: str
    ) -> None:
        """Save data as hdf5 file.

        If data is given as `dict[str, dict[str, Tensor]]`, the data will be saved as a group in the hdf5 file.

        Args:
            data (dict[str, Tensor] | dict[str, dict[str, Tensor]]): data to save
            file_name (str): File name. It can be either with or without .h5 extension.
        """
        log_dir = Path(self.save_dir, file_name)

        if log_dir.suffix == "":
            log_dir = log_dir.with_suffix(".h5")

        with h5py.File(log_dir, "w") as f:
            for k, v in data.items():
                if isinstance(v, dict):
                    grp = f.create_group(k)
                    for k1, v1 in v.items():
                        grp.create_dataset(
                            k1, shape=v1.shape, dtype=DTYPE_T2N[v1.dtype], data=v1
                        )
                else:
                    f.create_dataset(k, shape=v.shape, dtype=DTYPE_T2N[v.dtype], data=v)

    def save_csv(
        self,
        data: dict[str, Tensor | float | int],
        file_name: str,
        itr: int | None = None,
    ):
        """Save data as csv file. If `itr` is not `None`, the data will be appended to the file every `itr`. Designed for saving time series data.

        data (dict[str, Tensor]): Data to save.
        file_name (str): File name. It can be either with or without .csv extension.
        itr (int, optional): Iteration key. Defaults to None.
        """
        log_dir = Path(self.save_dir, file_name)

        if log_dir.suffix == "":
            log_dir = log_dir.with_suffix(".csv")

        # If file already exists, delete it
        if itr is None:
            pd_mode = "w"
            pd_header = True

            df = pd.DataFrame(data)
            df.to_csv(log_dir.as_posix(), header=pd_header, mode=pd_mode)

        else:
            if itr == 0:
                # If file already exists, delete it
                if is_file(log_dir):
                    log_dir.unlink()

                pd_mode = "w"
                pd_header = True
            else:
                # If itr is given as integer, append data to the file without header
                pd_mode = "a"
                pd_header = False

            data_to_save = {}
            for k, v in data.items():
                data_to_save[k] = [v]
            df = pd.DataFrame(data_to_save)
            df.to_csv(log_dir.as_posix(), header=pd_header, mode=pd_mode)

    def save_vtu(
        self,
        data: dict[str, Tensor] | list[dict[str, Tensor]],
        file_name: str | list[str],
        chunk: int | None = None,
        shuffle: bool = False,
    ) -> None:
        """Save VTK point (unstructured grid) data (.vtu format).

        The purpose of this class is to save particle data. Therefore, it only accepts particle position and velocity data.

        Note:
            - The data must be in the form of `dict` with keys `pos` and `vel`.

        Example:
            >>> pos_e = torch.tensor(...) # shape: num_particle x 3
            >>> vel_e = torch.tensor(...) # shape: num_particle x 3
            >>> ds = DataSaver(save_dir="./data")
            >>> ds.save_vtu(
                {"pos": pos_e, "vel": vel_e}, "p_e.vtu"
                ) # Save to ./data/p_e.vtu
            # Or you can save multiple data at once by passing list of dict and list of file name
            >>> pos_i = torch.tensor(...) # shape: num_particle x 3
            >>> vel_i = torch.tensor(...) # shape: num_particle x 3
            >>> ds.save_vtu(
                [{"pos": pos_e, "vel": vel_e}, {"pos": pos_i, "vel": vel_i}],
                ["p_e", "p_i"]
                ) # Save to ./data/p_e.vtu and ./data/p_i.vtu

        Args:
            data (dict[str, Tensor] | list[dict[str, Tensor]]): data to save. Either dictionary or list of dictionary.
            file_name (str | list[str]): file name to save. Either string or list of string.
            chunk (int | None, optional): chunk data. Save only `filter` number of particle to the file. Defaults to None (save all).
            shuffle (bool, optional): Shuffle data to save. Defaults to False.
        """

        if isinstance(data, dict):
            data = [data]

        if isinstance(file_name, str):
            file_name = [file_name]

        assert len(data) == len(
            file_name
        ), "Dataloader: dict_data and file_name must have the same length."

        for idx, d in enumerate(data):
            slicer = slice(0, chunk)
            pos = d["pos"]
            vel = d["vel"]

            # Shuffle data. Mostly used for saving chunk of data.
            # If not shuffled, only a few first particles in the tensor will be saved.
            # This is not an ideal way to save data in case there is inflow or outflow in the simulation. Therefore, we shuffle the data.
            if shuffle:
                n_tot = pos.shape[0]
                pos = pos[torch.randperm(n_tot), :]
                vel = vel[torch.randperm(n_tot), :]

            log_dir = Path(self.save_dir, file_name[idx])

            # Do I need detach and cpu?
            pyevtk.hl.pointsToVTK(
                path=log_dir.as_posix(),
                x=np.ascontiguousarray(pos[slicer, 0]),
                y=np.ascontiguousarray(pos[slicer, 1]),
                z=np.ascontiguousarray(pos[slicer, 2]),
                data={
                    "U": np.ascontiguousarray(vel[slicer, 0]),
                    "V": np.ascontiguousarray(vel[slicer, 1]),
                    "W": np.ascontiguousarray(vel[slicer, 2]),
                },
                fieldData=None,
            )

    def save_vti(
        self,
        grid: tuple[Tensor, ...],
        data: dict[str, Tensor],
        file_name: str,
        empty_dim_scale: list[float] = [1.0, 1.0],
    ) -> None:
        """Save gridded data.

        Note:
            - Data always saved in 3D. Therefore, if the data is 2D, we add one dimension to the data.
            - To save in 3D, we introduce empty_dim_scale for non 3D data. It places one grid point to the empty dimension with scale factor.

        Example:
            >>> grid = torch.meshgrid(...)
            >>> field1 = torch.tensor(...)
            >>> field2 = torch.tensor(...)
            >>> ds = DataSaver(save_dir="./data")
            >>> ds.save_vti(
                {"field1": field1, "field2": field2}, "field_data"
                ) # Save to ./data/field_data.vtk

        Args:
            grid (tuple[Tensor, ...]): Grid information. Should be the result of torch.meshgrid.
            data (dict[str, Tensor]): Data to save. Data in the dictionary should have same shape of grid entities.
            file_name (str): File name to save.
            empty_dim_scale (list[float], optional): Scale factor for empty dimension. Defaults to [1.0, 1.0].
        """

        # Get grid information
        dim = len(grid)
        nx = grid[0].shape
        dx = [
            ((grid[i].max() - grid[i].min()) / (nx[i] - 1)).item() for i in range(dim)
        ]
        lower = [grid[i].min().item() - 0.5 * dx[i] for i in range(dim)]

        to_save = {}

        # Assign dummy delta x for 1 and 2D cases
        if dim == 1:
            view_size = (*nx, 1, 1)

            g_min = (
                *lower,
                -dx[0] * empty_dim_scale[0],
                -dx[0] * empty_dim_scale[1],
            )
            g_dx = (
                *dx,
                dx[0] * empty_dim_scale[0],
                dx[0] * empty_dim_scale[1],
            )
        elif dim == 2:
            view_size = (*nx, 1)

            g_min = (
                *lower,
                -dx[0] * empty_dim_scale[0],
            )
            g_dx = (
                *dx,
                dx[0] * empty_dim_scale[0],
            )
        else:
            view_size = (*nx,)
            g_min = (*lower,)
            g_dx = (*dx,)

        for k, v in data.items():
            # Convert torch tensor to numpy array since pyevtk only accepts numpy array
            data_to_save = (v.view(*view_size)).detach().cpu().numpy()

            # If data dtype is booleon, convert to int
            if data_to_save.dtype == np.bool_:
                data_to_save = data_to_save.astype(np.int32)

            to_save.setdefault(k, data_to_save)

        log_dir = Path(self.save_dir, file_name)

        pyevtk.hl.imageToVTK(
            path=log_dir.as_posix(),
            origin=g_min,
            spacing=g_dx,
            cellData=to_save,
        )


def is_dir(log_dir: PATHLIKE, create: bool = True, verbose: int = 0) -> bool:
    """Check directory and if doesn't exist, create directory.

    Example:
        # No `./data` directory exists.
        >>> is_dir("./data", create=False, verbose=1) # Check if directory exists.
        >>> is_dir("./data", create=True, verbose=1) # Create directory if it doesn't exist.
        True
        >>> is_dir("./data", create=False, verbose=1) # Check if directory exists.
        True

    Args:
        log_dir (PATHLIKE): Directory to check.
        create (bool, optional): Create directory if it doesn't exist. Defaults to True.
        verbose (int, optional): Verbosity level. Defaults to 0.
    """
    from pymytools.logger import logging, markup

    log_dir = Path(log_dir).resolve()

    if log_dir.exists():
        if verbose > 0:
            logging.info(markup(f"Directory exists at {log_dir}", "green", "bold"))
        return True
    else:
        if create:
            if verbose > 0:
                logging.info(markup(f"Creating directory: {log_dir}", "yellow", "bold"))
            log_dir.mkdir(parents=True)
            return True
        else:
            logging.info(markup(f"Directory doesn't exist!", "red", "bold"))
            return False


def is_file(file_path: PATHLIKE) -> bool:
    """Check file and if doesn't exist, create file."""

    file_path = Path(file_path)

    return file_path.is_file()


class DataTracker:
    """Track simulation data using tensorboard."""

    stage: Enum | None

    def __init__(self, log_dir: PATHLIKE, overwrite: bool = True):
        self.log_dir = log_dir

        is_dir(self.log_dir, create=True)

        if overwrite:
            pass
        else:
            self.clear_data()

        self._writer = SummaryWriter(log_dir)

    def set_stage(self, stage: Enum):
        """Set stage (Enum class)."""

        self.stage = stage

    def add_batch_metric(self, name: str, value: float, step: int) -> None:
        """Add batch metric to tensorboard."""

        if self.stage is None:
            msg = f"batch/{name}"
        else:
            msg = f"{self.stage.name}/batch/{name}"

        self._writer.add_scalar(msg, value, step)

    def add_epoch_metric(self, name: str, value: float, step: int) -> None:
        """Add epoch metric to tensorboard."""

        if self.stage is None:
            msg = f"epoch/{name}"
        else:
            msg = f"{self.stage.name}/epoch/{name}"

        self._writer.add_scalar(msg, value, step)

    def add_scalar(self, msg: str, value: float, step: int) -> None:
        """Add scalar metric to tensorboard."""

        if self.stage is None:
            msg = msg
        else:
            msg = f"{self.stage.name}/{msg}"

        self._writer.add_scalar(msg, value, step)

    def add_scalars(self, msg: str, value_dict: dict, step: int) -> None:
        """Add scalar metric to tensorboard."""

        self._writer.add_scalars(msg, value_dict, step)

    def add_figure(self, msg: str, fig: Figure, step: int) -> None:
        """Add matplotlib figure to tensorboard."""

        self._writer.add_figure(msg, fig, step)

    def clear_data(self) -> None:
        """Clear generated data for the tensorboard."""

        # Remove files in the log directory
        pth = Path(self.log_dir)
        rm_tree(pth)

    def flush(self):
        """Flush the data."""

        self._writer.flush()


def rm_tree(pth: Path):
    """Recursively remove all files in the directory."""

    # Remove all files
    for child in pth.iterdir():
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    # Remove directory
    pth.rmdir()


def file_list_by_pattern(data_dir: PATHLIKE, pattern: str) -> list[str]:
    """Check directory and return list of all files matching the pattern.

    Example:
        # In `./data/` directory, there are `data_0000.h5`, `data_0001.h5`, `data_0002.h5`.
        >>> file_list_by_pattern("./data/", "data_*.h5")
        ['./data/data_0000.h5', './data/data_0001.h5', './data/data_0002.h5']
    """

    path = Path(data_dir, pattern).as_posix()

    # else if the directory is given, read all h5 files.
    file_list = glob.glob(path)

    return file_list



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pymytools', 'pymytools.special']

package_data = \
{'': ['*']}

install_requires = \
['h5py>=3.8.0,<4.0.0',
 'pandas>=1.5.3,<2.0.0',
 'pyevtk>=1.5.0,<2.0.0',
 'pymyplot>=0.2.7,<0.3.0',
 'rich>=13.3.2,<14.0.0',
 'scipy>=1.10.1,<2.0.0',
 'simpleaudio>=1.0.4,<2.0.0',
 'tensorboard>=2.12.0,<3.0.0',
 'torch>=2.0.0,<3.0.0',
 'tqdm>=4.65.0,<5.0.0',
 'vtk>=9.2.6,<10.0.0']

setup_kwargs = {
    'name': 'pymytools',
    'version': '0.1.14',
    'description': 'Tools used in my pystops project.',
    'long_description': '# Collection of Useful Tools for `pystops` Project\n\nDuring my PhD, I noticed that some features that I implemented throughout my projects appeared repeatedly.\nAs a result, I decided to separate these features and manage them appropriately.\n\n## Goal\n\nManage/unify fragmented features and deploy to `pypi.org`.\n\n## Features\n\n- Diagnostics\n  - file I/O\n    - Supports: `csv`, `hdf5` (`.h5`), and `vtk` (both `.vti` and `.vtu`) formats\n    - All data loaded are `torch.Tensor`\n  - Tensorboard tracker\n- Logging, timer, and progress bar\n  - Prettier representation using `rich` package\n\n## Dependencies\n\nMain dependencies:\n\n- `python = "^3.10"`\n- `h5py = "^3.8.0"`\n- `pyevtk = "^1.5.0"`\n- `rich = "^13.3.2"`\n- `tensorboard = "^2.12.0"`\n- `tqdm = "^4.65.0"`\n- `pandas = "^1.5.3"`\n- `torch = "^1.13.1"`\n- `vtk = "^9.2.6"`\n\nMy other personal project:\n\n- `pymyplot = "^0.2.7"` (for plotting)\n',
    'author': 'Kyoungseoun Chung',
    'author_email': 'kyoungseoun.chung@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)
